"""Middleware for the redis plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.redis

class RedisMiddleware(object):
    """Adds a redis connection to the RequestHandler"""

    def redis(self):
        """Gets or creates a connection to the redis database"""
        return oz.redis.create_connection()



"""The redis sessions plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import os
import binascii
import hashlib
from tornado import escape

from .middleware import *
from .options import *

def random_hex(length):
    """Generates a random hex string"""
    return escape.to_unicode(binascii.hexlify(os.urandom(length))[length:])

def password_hash(password, password_salt=None):
    """Hashes a specified password"""
    password_salt = password_salt or oz.settings["session_salt"]
    salted_password = password_salt + password
    return "sha256!%s" % hashlib.sha256(salted_password.encode("utf-8")).hexdigest()



"""UIModules for the error pages plugin"""

import oz
import base64
import pprint
import oz.error_pages
import tornado.web
import tornado.escape

TABLE_FORMAT = """
<table %s %s>
    <thead>
        <tr>
            <th>Variable</th>
            <th>Value</th>
        </tr>
    </thead>
    <tbody>
        %s
    </tbody>
</table>
"""

TABLE_ROW_FORMAT = """
<tr>
    <td>%s %s</td>
    <td class="code">%s</td>
</tr>
"""

@oz.uimodule
class DictTable(tornado.web.UIModule):
    """Renders an HTML table from a dict"""

    def render(self, d, id=None, kls=None):
        items = sorted(d.items())

        if items:
            rows = []

            for k, v in items:
                try:
                    escaped_val = tornado.escape.xhtml_escape(oz.error_pages.prettify_object(v))
                    rows.append(TABLE_ROW_FORMAT % (k, "", escaped_val))
                except UnicodeDecodeError:
                    rows.append(TABLE_ROW_FORMAT % (k, "(in base64)", base64.b64encode(v)))

            return TABLE_FORMAT % ("id='%s'" % id if id else "", "class='%s'" if kls else "", "\n".join(rows))
        else:
            return "<p>No data</p>"



"""Middleware for the sqlalchemy plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.sqlalchemy

class SQLAlchemyMiddleware(object):
    """Adds a per-request sqlalchemy transaction"""

    def __init__(self):
        super(SQLAlchemyMiddleware, self).__init__()
        self.trigger_listener("on_finish", self._sqlalchemy_on_finish)

    def db(self):
        """Gets the SQLALchemy session for this request"""

        if not hasattr(self, "db_conn"):
            self.db_conn = oz.sqlalchemy.session()

        return self.db_conn

    def _sqlalchemy_on_finish(self):
        """
        Closes the sqlalchemy transaction. Rolls back if an error occurred.
        """

        if hasattr(self, "db_conn"):
            if self.get_status() >= 200 and self.get_status() <= 399:
                self.db_conn.commit()
            else:
                self.db_conn.rollback()

            self.db_conn.close()



"""Oz"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import types
import collections
import tornado.web
import tornado.options
import tornado.util
import tornado.log
import inspect
import functools
import os

# On trigger execution, trigger listeners can return this to notify the
# request handler to cancel execution of the next functions in the trigger
# chain.
break_trigger = object()

# Mapping of action name -> callback
_actions = {}

# Mapping of uimodule name -> class
_uimodules = {}

# List of routes
_routes = []

# List of test classes
_tests = []

# Mapping of setting name -> value
settings = {}

def _add_to_dict(t, container, name, value):
    """
    Adds an item to a dictionary, or raises an exception if an item with the
    specified key already exists in the dictionary.
    """

    if name in container:
        raise Exception("%s '%s' already exists" % (t, name))
    else:
        container[name] = value

def action(fun):
    """Exposes an action"""
    _add_to_dict("Action", _actions, fun.__name__, fun)
    return fun

def uimodule(cls):
    """Exposes a UIModule"""
    _add_to_dict("UIModules", _uimodules, cls.__name__, cls)
    return cls

def route(new_route):
    """Exposes a route"""
    _routes.append(new_route)

def routes(*new_routes):
    """Exposes a list of routes"""
    _routes.extend(new_routes)

def option(name, **args):
    """Exposes an option"""
    tornado.options.define(name, **args)

def options(**kwargs):
    """Exposes several options"""
    for name, args in kwargs.items():
        option(name, **args)

def test(cls):
    """Exposes a unit test class, to be run on the `test` action."""
    _tests.append(cls)
    return cls

def plugin(namespace):
    """Loads an oz plugin"""
    __import__(namespace, globals(), locals(), [], 0)

class RequestHandler(tornado.web.RequestHandler):
    """
    Builds on top of tornado's RequestHandler to provide template helpers and triggers
    """

    def __init__(self, *args, **kwargs):
        self._template_helpers = {}
        self._triggers = collections.defaultdict(lambda: [])
        super(RequestHandler, self).__init__(*args, **kwargs)

    def template_helper(self, name, callback):
        """Adds a template helper"""
        _add_to_dict("Template helper", self._template_helpers, name, callback)

    def trigger_listener(self, name, callback):
        """Adds a trigger listener"""
        self._triggers[name].append(callback)

    def trigger(self, name, *args, **kwargs):
        """
        Triggers an event to run through middleware. This method will execute
        a chain of relevant trigger callbacks, until one of the callbacks
        returns the `break_trigger`.
        """

        # Relevant middleware is cached so we don't have to rediscover it
        # every time. Fetch the cached value if possible.

        listeners = self._triggers.get(name, [])

        # Execute each piece of middleware
        for listener in listeners:
            result = listener(*args, **kwargs)

            if result == break_trigger:
                return False

        return True

    def get_template_namespace(self):
        # Start with the default namespace
        namespace = super(RequestHandler, self).get_template_namespace()

        # Add the template helpers
        namespace.update(self._template_helpers)

        return namespace

    def initialize(self, **kwargs):
        self.trigger("initialize", **kwargs)

    def set_secure_cookie(self, name, value, **kwargs):
        if self.settings["use_secure_cookie"]:
            if "secure" not in kwargs:
                kwargs["secure"] = True
            if "httponly" not in kwargs:
                kwargs["httponly"] = True
        super(RequestHandler, self).set_secure_cookie(name, value, **kwargs)

    def prepare(self):
        if self.settings["use_hsts"]:
            self.set_header("Strict-Transport-Security", "max-age=2592000; includeSubDomains")

        self.trigger("prepare")

    def on_finish(self):
        self.trigger("on_finish")

    def on_connection_close(self):
        self.trigger("on_connection_close")

    def write_error(self, *args, **kwargs):
        # Call super's implementation of `write_error` unless we hit a
        # `break_trigger`.
        if self.trigger("write_error", *args, **kwargs):
            super(RequestHandler, self).write_error(*args, **kwargs)

def initialize(config=None):
    """Initializes oz"""

    # Load the config file
    if config == None:
        config = {}
        config_source = None

        try:
            with open(os.environ.get("OZ_CONFIG", "config.py")) as f:
                config_source = f.read()
        except Exception:
            tornado.log.gen_log.info("Could not read config.py", exc_info=True)

        if config_source != None:
            tornado.util.exec_in(config_source, config, config)

    # Load the plugins
    for p in config.get("plugins", ["oz.core"]):
        plugin(p)

    # Set the options
    for key, value in config.get("app_options", {}).items():
        setattr(tornado.options.options, key, value)

    # Generate the application settings
    global settings
    settings = tornado.options.options.as_dict()
    settings["ui_modules"] = _uimodules
    settings["project_name"] = config.get("project_name")



"""Middleware for the error pages plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.error_pages
import sys

class ErrorPageMiddleware(object):
    """Middleware to show the custom error page"""

    def __init__(self):
        super(ErrorPageMiddleware, self).__init__()
        self.trigger_listener("write_error", self._on_error_page_write_error)

    def _on_error_page_write_error(self, status_code, **kwargs):
        """Replaces the default Tornado error page with a Django-styled one"""

        if oz.settings.get('debug'):
            exception_type, exception_value, tback = sys.exc_info()
            is_breakpoint = isinstance(exception_value, oz.error_pages.DebugBreakException)

            frames = oz.error_pages.get_frames(tback, is_breakpoint)
            frames.reverse()

            if is_breakpoint:
                exception_type = 'Debug breakpoint'
                exception_value = ''

            self.render(oz.settings["error_pages_template"],
                exception_type=exception_type,
                exception_value=exception_value,
                frames=frames,
                request_input=self.request.body,
                request_cookies=self.cookies,
                request_headers=self.request.headers,
                request_path=self.request.uri,
                request_method=self.request.method,
                response_output="".join(self._write_buffer),
                response_headers=self._headers,
                prettify_object=oz.error_pages.prettify_object,
            )

            return oz.break_trigger



"""Middleware for the redis sessions plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import oz.redis_sessions

class RedisSessionMiddleware(object):
    """Adds redis-backed session capabilities"""

    @property
    def _session_key(self):
        """Gets the redis key for a session"""

        if not hasattr(self, "_cached_session_key"):
            session_id_bytes = self.get_secure_cookie("session_id")
            session_id = None

            if session_id_bytes:
                try:
                    session_id = session_id_bytes.decode('utf-8')
                except:
                    pass

            if not session_id:
                session_id = oz.redis_sessions.random_hex(20)
                self.set_secure_cookie(
                    name="session_id",
                    value=session_id.encode('utf-8'),
                    domain=oz.settings.get("cookie_domain"),
                    httponly=True)

            password_salt = oz.settings["session_salt"]
            self._cached_session_key = "session:%s:v4" % oz.redis_sessions.password_hash(session_id, password_salt=password_salt)

        return self._cached_session_key

    def _update_session_expiration(self):
        """
        Updates a redis item to expire later since it has been interacted with
        recently
        """

        session_time = oz.settings["session_time"]

        if session_time:
            self.redis().expire(self._session_key, session_time)

    def get_session_value(self, name, default=None):
        """Gets a session value"""

        value = self.redis().hget(self._session_key, name) or default
        self._update_session_expiration()
        return value

    def set_session_value(self, name, value):
        """Sets a session value"""

        self.redis().hset(self._session_key, name, value)
        self._update_session_expiration()

    def clear_session_value(self, name):
        """Removes a session value"""
        self.redis().hdel(self._session_key, name)
        self._update_session_expiration()

    def clear_all_session_values(self):
        """Kills a session"""
        self.redis().delete(self._session_key)
        self.clear_cookie("session_id")



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import datetime
import oz.redis
import oz.bandit
from tornado import escape

@oz.action
def add_experiment(experiment):
    """Adds a new experiment"""
    redis = oz.redis.create_connection()
    oz.bandit.add_experiment(redis, experiment)

@oz.action
def archive_experiment(experiment):
    """Archives an experiment"""
    redis = oz.redis.create_connection()
    oz.bandit.Experiment(redis, experiment).archive()

@oz.action
def add_experiment_choice(experiment, choice):
    """Adds an experiment choice"""
    redis = oz.redis.create_connection()
    oz.bandit.Experiment(redis, experiment).add_choice(choice)

@oz.action
def remove_experiment_choice(experiment, choice):
    """Removes an experiment choice"""
    redis = oz.redis.create_connection()
    oz.bandit.Experiment(redis, experiment).remove_choice(choice)

@oz.action
def get_experiment_results():
    """
    Computes the results of all experiments, stores it in redis, and prints it
    out
    """

    redis = oz.redis.create_connection()

    for experiment in oz.bandit.get_experiments(redis):
        experiment.compute_default_choice()
        csq, confident = experiment.confidence()

        print("%s:" % experiment.name)
        print("- creation date: %s" % experiment.metadata["creation_date"])
        print("- default choice: %s" % experiment.default_choice)
        print("- chi squared: %s" % csq)
        print("- confident: %s" % confident)
        print("- choices:")

        for choice in experiment.choices:
            print("  - %s: plays=%s, rewards=%s, performance=%s" % (choice.name, choice.plays, choice.rewards, choice.performance))

@oz.action
def sync_experiments_from_spec(filename):
    """
    Takes the path to a JSON file declaring experiment specifications, and
    modifies the experiments stored in redis to match the spec.

    A spec looks like this:
    {
       "experiment 1": ["choice 1", "choice 2", "choice 3"],
       "experiment 2": ["choice 1", "choice 2"]
    }
    """

    redis = oz.redis.create_connection()

    with open(filename, "r") as f:
        schema = escape.json_decode(f.read())

    oz.bandit.sync_from_spec(redis, schema)



"""The error pages plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from .middleware import *
from .uimodules import *
from .options import *

import traceback
import base64
import pprint
import linecache

class DebugBreakException(Exception):
    """Raise this to break into the debugger during an HTTP request"""
    pass

def debug():
    """Used to create debug breakpoints in code"""
    raise DebugBreakException()

class ErrorFrame(object):
    """Holds information about a function call in a traceback"""

    def __init__(self, tback, filename, function, lineno, vars, id, pre_context, context_line, post_context, pre_context_lineno):
        self.tback = tback
        self.filename = filename
        self.function = function
        self.lineno = lineno
        self.vars = vars
        self.id = id
        self.pre_context = pre_context
        self.context_line = context_line
        self.post_context = post_context
        self.pre_context_lineno = pre_context_lineno

def get_lines_from_file(filename, lineno, context_lines):
    """
    Returns context_lines before and after lineno from file.
    Returns (pre_context_lineno, pre_context, context_line, post_context).
    """

    def get_lines(start, end):
        return [linecache.getline(filename, l).rstrip() for l in range(start, end)]

    lower_bound = max(1, lineno - context_lines)
    upper_bound = lineno + context_lines
    linecache.checkcache(filename)
    pre_context = get_lines(lower_bound, lineno)
    context_line = linecache.getline(filename, lineno).rstrip()
    post_context = get_lines(lineno + 1, upper_bound)
    return lower_bound, pre_context, context_line, post_context

def get_frames(tback, is_breakpoint):
    """Builds a list of ErrorFrame objects from a traceback"""

    frames = []

    while tback is not None:
        if tback.tb_next is None and is_breakpoint:
            break

        filename = tback.tb_frame.f_code.co_filename
        function = tback.tb_frame.f_code.co_name
        context = tback.tb_frame.f_locals
        lineno = tback.tb_lineno - 1
        tback_id = id(tback)
        pre_context_lineno, pre_context, context_line, post_context = get_lines_from_file(filename, lineno + 1, 7)
        frames.append(ErrorFrame(tback, filename, function, lineno, context, tback_id, pre_context, context_line, post_context, pre_context_lineno))
        tback = tback.tb_next

    return frames

def prettify_object(obj):
    """Makes a pretty string for an object for nice output"""

    try:
        return pprint.pformat(str(obj))
    except UnicodeDecodeError as e:
        raise
    except Exception as e:
        return "[could not display: <%s: %s>]" % (e.__class__.__name__, str(e))



"""Actions for the core plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import wsgiref.simple_server
import signal
import time
import sys
import shutil
import unittest
import os
import functools
import re

import tornado.web
import tornado.wsgi
import tornado.httpserver
import tornado.ioloop
import tornado.log

import oz

VALID_PROJECT_NAME = re.compile(r"^\w+$")

def check_path(path, otherwise):
    """
    Checks if a path exists. If it does, print a warning message; if not,
    execute the `otherwise` callback argument.
    """

    if os.path.exists(path):
        print("WARNING: Path '%s' already exists; skipping" % path)
    else:
        otherwise(path)

def config_maker(project_name, path):
    """Creates a config file based on the project name"""

    with open(skeleton_path("config.py"), "r") as config_source:
        config_content = config_source.read()

    config_content = config_content.replace("__PROJECT_NAME__", project_name)

    with open(path, "w") as config_dest:
        config_dest.write(config_content)

def skeleton_path(parts):
    """Gets the path to a skeleton asset"""
    return os.path.join(os.path.dirname(oz.__file__), "skeleton", parts)

@oz.action
def init(project_name):
    """Creates a new project"""

    if not VALID_PROJECT_NAME.match(project_name):
        print("Invalid project name. It may only contain letters, numbers and underscores.", file=sys.stderr)
        return

    check_path(project_name, functools.partial(shutil.copytree, skeleton_path("plugin")))
    check_path("static", os.mkdir)
    check_path("templates", os.mkdir)
    check_path("config.py", functools.partial(config_maker, project_name))

@oz.action
def server():
    """Runs the server"""

    # Get and validate the server_type
    server_type = oz.settings["server_type"]
    if server_type not in [None, "wsgi", "asyncio", "twisted"]:
        raise Exception("Unknown server type: %s" % server_type)

    # Install the correct ioloop if necessary
    if server_type == "asyncio":
        from tornado.platform.asyncio import AsyncIOMainLoop
        AsyncIOMainLoop().install()
    elif server_type == "twisted":
        from tornado.platform.twisted import TwistedIOLoop
        TwistedIOLoop().install()

    if server_type == "wsgi":
        wsgi_app = tornado.wsgi.WSGIApplication(oz._routes, **oz.settings)
        wsgi_srv = wsgiref.simple_server.make_server("", oz.settings["port"], wsgi_app)
        wsgi_srv.serve_forever()
    else:
        web_app = tornado.web.Application(oz._routes, **oz.settings)

        if oz.settings["ssl_cert_file"] != None and oz.settings["ssl_key_file"] != None:
            ssl_options = {
                "certfile": oz.settings["ssl_cert_file"],
                "keyfile": oz.settings["ssl_key_file"],
                "cert_reqs": oz.settings["ssl_cert_reqs"],
                "ca_certs": oz.settings["ssl_ca_certs"]
            }
        else:
            ssl_options = None

        http_srv = tornado.httpserver.HTTPServer(
            web_app,
            ssl_options=ssl_options,
            body_timeout=oz.settings["body_timeout"],
            xheaders=oz.settings["xheaders"]
        )

        http_srv.bind(oz.settings["port"])

        if oz.settings["debug"]:
            if oz.settings["server_workers"] != 1:
                print("WARNING: Debug is enabled, but multiple server workers have been configured. Only one server worker can run in debug mode.")

            http_srv.start(1)
        else:
            # Forks multiple sub-processes
            http_srv.start(oz.settings["server_workers"])

        if oz.settings.get("use_graceful_shutdown"):
            if server_type == "asyncio" or server_type == "twisted":
                print("WARNING: Cannot enable graceful shutdown for asyncio or twisted server types.")
            else:
                # NOTE: Do not expect any logging to with certain tools (e.g., invoker),
                # because they may quiet logs on SIGINT/SIGTERM
                signal.signal(signal.SIGTERM, functools.partial(_shutdown_tornado_ioloop, http_srv))
                signal.signal(signal.SIGINT, functools.partial(_shutdown_tornado_ioloop, http_srv))

        if server_type == "asyncio":
            import asyncio
            asyncio.get_event_loop().run_forever()
        elif server_type == "twisted":
            from twisted.internet import reactor
            reactor.run()
        else:
            from tornado import ioloop
            ioloop.IOLoop.instance().start()

@oz.action
def repl():
    """Runs an IPython repl with some context"""

    try:
        import IPython
    except:
        print("ERROR: IPython is not installed. Please install it to use the repl.", file=sys.stderr)
        raise

    IPython.embed(user_ns=dict(
        settings=oz.settings,
        actions=oz._actions,
        uimodules=oz._uimodules,
        routes=oz._routes,
    ))

@oz.action
def test(*filters):
    """Runs the unit tests"""

    suite = unittest.TestSuite()
    filters_set = set(filters)

    for t in oz._tests:
        if not filters_set or t.__name__ in filters_set:
            # Add the class to the suite
            child_suite = unittest.makeSuite(t, "test")
            suite.addTest(child_suite)

    res = unittest.TextTestRunner().run(suite)
    return 1 if len(res.errors) > 0 or len(res.failures) > 0 else 0

def _shutdown_tornado_ioloop(http_srv, sig, frame):
    io_loop = tornado.ioloop.IOLoop.instance()

    def stop_loop(deadline_seconds):
        now = time.time()
        if now < deadline_seconds and io_loop._callbacks:
            io_loop.add_timeout(now + 1, stop_loop, deadline_seconds)
        else:
            io_loop.stop()

    def shutdown():
        tornado.log.app_log.info("Shutting down server....")
        http_srv.stop()
        stop_loop(time.time() + oz.settings.get("graceful_shutdown_timeout", 5))

    tornado.log.app_log.warning("Received %s", sig)
    io_loop.add_callback_from_signal(shutdown)



"""Handles CLI functionality for oz"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
import optfn
import sys

def main():
    """Main entry-point for oz's cli"""

    # Hack to make user code available for import
    sys.path.append(".")

    # Run the specified action
    oz.initialize()
    retr = optfn.run(list(oz._actions.values()))

    if retr == optfn.ERROR_RETURN_CODE:
        sys.exit(-1)
    elif retr == None:
        sys.exit(0)
    elif isinstance(retr, int):
        sys.exit(retr)
    else:
        raise Exception("Unexpected return value from action: %s" % retr)



"""Options for the sqlalchemy plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    db = dict(type=str, help="SQLAlchemy connection string to the database"),
    debug_sql = dict(type=bool, default=False, help="Dump SQL queries performed by SQLAlchemy to stdout"),
    db_pool_size = dict(type=int, default=None, help="Size of the SQLAlchemy connection pool"),
    db_max_overflow = dict(type=int, default=None, help="How many extra transient connections can be put in the SQLAlchemy connection pool if needed"),
    db_pool_timeout = dict(type=int, default=None, help="Number of seconds to wait before a SQLAlchemy connection is returned to the pool"),
)


"""Options for the core plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    debug = dict(type=bool, default=False, help="Debug mode. Stack traces will be displayed on errors, and the server will automatically reload when server code is changed. This should not be enabled in production."),
    port = dict(type=int, default=8000, help="The port on which to run the server."),
    static_path = dict(type=str, default=None, help="Path to the static directory. If set to None, then static asset hosting is disabled."),
    template_path = dict(type=str, default="templates", help="Path to the templates directory."),
    xsrf_cookies = dict(type=bool, default=False, help="Enable cookies that prevent XSRF attacks."),
    cookie_secret = dict(type=str, help="Secret used for secure cookies. Required if you set xsrf_cookies=True."),
    gzip = dict(type=bool, help="Enables gzip output."),
    wsgi_mode = dict(type=bool, default=False, help="Use WSGI mode."),
    server_workers = dict(type=int, default=1, help="The number of server workers to run concurrently."),
    body_timeout = dict(type=int, default=None, help="How long to wait in seconds while reading the request body"),
    xheaders = dict(type=bool, default=False, help="Determine remote_ip from X-Remote-Ip/X-Forwarded-For headers"),
    ssl_cert_file = dict(type=str, default=None, help="The SSL certificate file. If unspecified, SSL support will be disabled."),
    ssl_key_file = dict(type=str, default=None, help="The SSL key file. If unspecified, SSL support will be disabled."),
    ssl_ca_certs = dict(type=str, default=None, help="SSL CA certificates file."),
    ssl_cert_reqs = dict(type=int, default=0, help="Whether certificates are required from the other side of the connection. 0 = certificates ignored, 1 = certificates optional, 2 = certificates required."),
    server_type = dict(type=str, default=None, help="By default, tornado will be started with the default ioloop. If this is set, a custom server can be used instead. Valid options: 'asyncio', 'twisted' or 'wsgi'."),
    use_secure_cookie=dict(type=bool, default=False, help="Set secure cookies with the secure flag"),
    use_hsts=dict(type=bool, default=False, help="Set the Strict-Transport-Security header"),
    use_graceful_shutdown=dict(type=bool, default=True, help="Stop the server gracefully"),
    graceful_shutdown_timeout=dict(type=int, default=5, help="Number of seconds to wait before forcing the server to stop")
)



"""Options for the aws_cdn plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    aws_access_key = dict(type=str, help="AWS access key for CDN"),
    aws_secret_key = dict(type=str, help="AWS secret key for CDN"),
    static_host = dict(type=str, help="CDN hostname for static assets"),
    s3_bucket = dict(type=str, default=None, help="S3 bucket for uploading CDN assets"),
)



import oz
import binascii
import tornado.web
from tornado.escape import xhtml_escape

try:
    from urllib.parse import urlparse, parse_qs
except ImportError:
    from urlparse import urlparse, parse_qs

class Subresource(tornado.web.UIModule):
    def get_integrity(self, url):
        parsed = urlparse(url)
        query = parse_qs(parsed.query)

        # cdn_static_url currently appends ?v=None to the URL if there is no
        # known hash of the. Default to this if `v` is somehow not defined as
        # well.
        v_arg = query.get("v", ["None"])[0]

        if v_arg == "None":
            return None
        else:
            # v_arg contains the hash in hex encoding - we need to convert it
            # to base64. This roundabout method should work in both python 2
            # 3.
            return "sha256-%s" % binascii.b2a_base64(bytearray.fromhex(v_arg)).strip().decode("utf8")

@oz.uimodule
class Script(Subresource):
    def render(self, path, sip=True, crossorigin="anonymous", **attrs):
        url = self.handler.cdn_static_url(path)

        if sip:
            integrity = self.get_integrity(url)

            if integrity and "integrity" not in attrs:
                attrs["integrity"] = integrity
            
            if "crossorigin" not in attrs:
                attrs["crossorigin"] = crossorigin

        if "src" not in attrs:
            attrs["src"] = url

        attr_strs = ['%s="%s"' % (xhtml_escape(k), xhtml_escape(v)) for (k, v) in attrs.items()]
        return "<script %s></script>" % " ".join(attr_strs)

@oz.uimodule
class Stylesheet(Subresource):
    def render(self, path, sip=True, crossorigin="anonymous", **attrs):
        url = self.handler.cdn_static_url(path)

        if sip:
            integrity = self.get_integrity(url)
            
            if integrity and "integrity" not in attrs:
                attrs["integrity"] = integrity
            
            if "crossorigin" not in attrs:
                attrs["crossorigin"] = crossorigin

        if "rel" not in attrs:
            attrs["rel"] = "stylesheet"
        if "href" not in attrs:
            attrs["href"] = url

        attr_strs = ['%s="%s"' % (xhtml_escape(k), xhtml_escape(v)) for (k, v) in attrs.items()]
        return "<link %s />" % " ".join(attr_strs)



"""Options for the error pages middleware"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz

oz.options(
    error_pages_template = dict(type=str, default="error_page.html", help="Special error page to render when an error occurs and debug is True. This should come from the error_pages plugin source code."),
)



"""The sqlalchemy plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from multiprocessing.util import register_after_fork
import tornado.options
import oz

from .actions import *
from .middleware import *
from .options import *

Base = declarative_base()
_engine = None
_session = None

class _AfterFork(object):
    """
    Ensures cleanup of sqlalchemy connections after a fork is done. This guard
    is here because sqlalchemy connections cannot be shared across processes.
    """

    registered = False

    def __call__(self):
        global _engine
        global _session

        # Child must reregister
        self.registered = False

        _engine.dispose()
        _engine = None
        _session = None

after_fork = _AfterFork()

def setup():
    """Initializes the tables if they don't exist already"""
    return Base.metadata.create_all(engine())

def engine():
    global _engine

    if _engine == None:
        kwargs = dict(echo=oz.settings["debug_sql"])

        if oz.settings["db_pool_size"]:
            kwargs["pool_size"] = oz.settings["db_pool_size"]
        if oz.settings["db_max_overflow"]:
            kwargs["max_overflow"] = oz.settings["db_max_overflow"]
        if oz.settings["db_pool_timeout"]:
            kwargs["pool_timeout"] = oz.settings["db_pool_timeout"]

        _engine = create_engine(oz.settings["db"], **kwargs)
        after_fork.registered = True
        register_after_fork(after_fork, after_fork)

    return _engine

def session():
    """Gets a SQLAlchemy session"""
    global _session

    if _session == None:
        _session = sessionmaker(bind=engine())

    return _session()



"""Actions for aws_cdn plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

# Module for generating hashes for files that match a glob, and putting that
# hash in redis to allow us to generate cache-busting URLs later

import os
import oz
import oz.redis
import oz.aws_cdn

@oz.action
def cache_busting_scan(*prefixes):
    """
    (Re-)generates the cache buster values for all files with the specified
    prefixes.
    """

    redis = oz.redis.create_connection()
    pipe = redis.pipeline()

    # Get all items that match any of the patterns. Put it in a set to
    # prevent duplicates.
    if oz.settings["s3_bucket"]:
        bucket = oz.aws_cdn.get_bucket()
        matches = set([oz.aws_cdn.S3File(key) for prefix in prefixes for key in bucket.list(prefix)])
    else:
        matches = set([])
        static_path = oz.settings["static_path"]
        
        for root, _, filenames in os.walk(static_path):
            for filename in filenames:
                path = os.path.relpath(os.path.join(root, filename), static_path)

                for prefix in prefixes:
                    if path.startswith(prefix):
                        matches.add(oz.aws_cdn.LocalFile(static_path, path))
                        break

    # Set the cache busters
    for f in matches:
        file_hash = f.hash()
        print(file_hash, f.path())
        oz.aws_cdn.set_cache_buster(pipe, f.path(), file_hash)

    pipe.execute()



"""Errors for the JSON API plugin"""

import tornado.web

class ApiError(tornado.web.HTTPError):
    """Exception for API-based errors"""

    def __init__(self, message, code=400):
        super(ApiError, self).__init__(code)
        self.message = message



from setuptools import setup

VERSION = "0.8.0"

setup(
    name="Oz",
    version=VERSION,
    description="A batteries-included web framework built on Tornado",
    author="Yusuf Simonson",
    author_email="yusuf@themuse.com",
    url="http://github.com/dailymuse/oz",
    zip_safe=False,

    packages=[
        "oz",
        "oz.aws_cdn",
        "oz.bandit",
        "oz.blinks",
        "oz.core",
        "oz.error_pages",
        "oz.json_api",
        "oz.redis",
        "oz.redis_sessions",
        "oz.sqlalchemy",
        "oz.error_pages"
    ],

    package_data={
        "oz": [
            "skeleton/*.py",
            "skeleton/plugin/*.py",
            "skeleton/plugin/tests/*.py",
        ]
    },

    install_requires=[
        "tornado>=3.1",
        "optfn>=0.4.1"
    ],

    extras_require={
        "oz.aws_cdn": ["boto>=2.9.7"],
        "oz.redis": ["redis>=2.6.0"],
        "oz.sqlalchemy": ["sqlalchemy>=0.7.8"]
    },

    entry_points={
        "console_scripts": ["oz = oz.cli:main"]
    }
)



"""Bandit testing plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

from tornado import escape, util
from .actions import *
from .middleware import *

import re

ACTIVE_EXPERIMENTS_REDIS_KEY = "bandit:listing:active:v2"
ARCHIVED_EXPERIMENTS_REDIS_KEY = "bandit:listing:archived:v2"
EXPERIMENT_REDIS_KEY_TEMPLATE = "bandit:experiment:%s:v2"
ALLOWED_NAMES = re.compile("^[A-Za-z0-9-_]+$")

# Via http://passel.unl.edu/Image/Namuth-CovertDeana956176274/chi-sqaure%20distribution%20table.PNG
# Ideally this would be computed on-the-fly. If you know your differential
# equations, PR please!
CHI_SQUARE_DISTRIBUTION = [
    3.84,
    5.99,
    7.81,
    9.49,
    11.07,
    12.59,
    14.07,
    15.51,
    16.92,
    18.31,
    19.68,
    21.03,
    22.36,
    23.68,
    25.00,
    26.30,
    27.59,
    28.87,
    30.14,
    31.41,
]

def chi_squared(*choices):
    """Calculates the chi squared"""

    term = lambda expected, observed: float((expected - observed) ** 2) / max(expected, 1)
    mean_success_rate = float(sum([c.rewards for c in choices])) / max(sum([c.plays for c in choices]), 1)
    mean_failure_rate = 1 - mean_success_rate

    return sum([
        term(mean_success_rate * c.plays, c.rewards)
        + term(mean_failure_rate * c.plays, c.plays - c.rewards
    ) for c in choices])

def is_confident(csq, num_choices):
    """
    Returns whether an experiment is statistically significant with 95%%
    confidence
    """
    return csq >= CHI_SQUARE_DISTRIBUTION[num_choices - 2]

def parse_json(raw):
    """Parses raw bytes to a JSON object with unicode strings"""
    return escape.recursive_unicode(escape.json_decode(raw)) if raw != None else None

def sync_from_spec(redis, schema):
    """
    Takes an input experiment spec and creates/modifies/archives the existing
    experiments to match the spec.

    If there's an experiment in the spec that currently doesn't exist, it will
    be created along with the associated choices.

    If there's an experiment in the spec that currently exists, and the set of
    choices are different, that experiment's choices will be modified to match
    the spec.

    If there's an experiment not in the spec that currently exists, it will be
    archived.

    A spec looks like this:

    {
       "experiment 1": ["choice 1", "choice 2", "choice 3"],
       "experiment 2": ["choice 1", "choice 2"]
    }
    """

    def get_experiments_dict(active=True):
        """Returns a dictionary of experiment names -> experiment objects"""
        return dict((experiment.name, experiment) for experiment in get_experiments(redis, active=active))

    # Get the current experiments
    active_experiments = get_experiments_dict()
    archived_experiments = get_experiments_dict(active=False)

    # Get the newly defined experiment names and the names of the experiments
    # already setup
    new_experiment_names = set(schema.keys())
    active_experiment_names = set(active_experiments.keys())

    # Find all the experiments that are in the schema and are defined among the
    # archived experiments, but not the active ones (we check against active
    # experiments to prevent the edge case where an experiment might be defined
    # doubly in both active and archived experiments)
    unarchivable_experiment_names = (new_experiment_names - active_experiment_names) & set(archived_experiments.keys())

    # De-archive the necessary experiments
    for unarchivable_experiment_name in unarchivable_experiment_names:
        print("- De-archiving %s" % unarchivable_experiment_name)

        # Because there is no function to de-archive an experiment, it must
        # be done manually
        pipe = redis.pipeline(transaction=True)
        pipe.sadd(ACTIVE_EXPERIMENTS_REDIS_KEY, unarchivable_experiment_name)
        pipe.srem(ARCHIVED_EXPERIMENTS_REDIS_KEY, unarchivable_experiment_name)
        pipe.execute()

    # Reload the active experiments if we de-archived any
    if unarchivable_experiment_names:
        active_experiments = get_experiments_dict()
        active_experiment_names = set(active_experiments.keys())

    # Create the new experiments
    for new_experiment_name in new_experiment_names - active_experiment_names:
        print("- Creating experiment %s" % new_experiment_name)

        experiment = add_experiment(redis, new_experiment_name)

        for choice in schema[new_experiment_name]:
            print("  - Adding choice %s" % choice)
            experiment.add_choice(choice)

    # Archive experiments not defined in the schema
    for archivable_experiment_name in active_experiment_names - new_experiment_names:
        print("- Archiving %s" % archivable_experiment_name)
        active_experiments[archivable_experiment_name].archive()

    # Update the choices for existing experiments that are also defined in the
    # schema
    for experiment_name in new_experiment_names & active_experiment_names:
        experiment = active_experiments[experiment_name]
        new_choice_names = set(schema[experiment_name])
        old_choice_names = set(experiment.choice_names)

        # Add choices in the schema that do not exist yet
        for new_choice_name in new_choice_names - old_choice_names:
            print("- Adding choice %s to existing experiment %s" % (new_choice_name, experiment_name))
            experiment.add_choice(new_choice_name)

        # Remove choices that aren't in the schema
        for removable_choice_name in old_choice_names - new_choice_names:
            print("- Removing choice %s from existing experiment %s" % (removable_choice_name, experiment_name))
            experiment.remove_choice(removable_choice_name)

class ExperimentException(Exception):
    """
    An exception for issues related to experiments. This way applications can
    catch the exception and respond accordingly.
    """

    def __init__(self, experiment_name, message):
        super(ExperimentException, self).__init__("Experiment %s: %s" % (experiment_name, message))
        self.experiment_name = experiment_name

class Experiment(object):
    """Specification for a bandit experiment"""

    def __init__(self, redis, name):
        self.redis = redis
        self.name = name
        self.refresh()

    def refresh(self):
        """Re-pulls the data from redis"""
        pipe = self.redis.pipeline()
        pipe.hget(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "metadata")
        pipe.hget(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "choices")
        pipe.hget(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "default-choice")
        results = pipe.execute()

        if results[0] == None:
            raise ExperimentException(self.name, "Does not exist")

        self.metadata = parse_json(results[0])
        self.choice_names = parse_json(results[1]) if results[1] != None else []
        self.default_choice = escape.to_unicode(results[2])
        self._choices = None

    @property
    def choices(self):
        """Gets the experiment choices"""

        if self._choices == None:
            self._choices = [ExperimentChoice(self, choice_name) for choice_name in self.choice_names]

        return self._choices

    def confidence(self):
        """
        Returns a tuple (chi squared, confident) of the experiment. Confident
        is simply a boolean specifying whether we're > 95%% sure that the
        results are statistically significant.
        """

        choices = self.choices

        # Get the chi-squared between the top two choices, if more than two choices exist
        if len(choices) >= 2:
            csq = chi_squared(*choices)
            confident = is_confident(csq, len(choices)) if len(choices) <= 10 else None
        else:
            csq = None
            confident = False

        return (csq, confident)

    def archive(self):
        """Archives an experiment"""
        pipe = self.redis.pipeline(transaction=True)
        pipe.srem(ACTIVE_EXPERIMENTS_REDIS_KEY, self.name)
        pipe.sadd(ARCHIVED_EXPERIMENTS_REDIS_KEY, self.name)
        pipe.execute()

    def add_choice(self, choice_name):
        """Adds a choice for the experiment"""

        if not ALLOWED_NAMES.match(choice_name):
            raise ExperimentException(self.name, "Illegal choice name: %s" % choice_name)

        if choice_name in self.choice_names:
            raise ExperimentException(self.name, "Choice already exists: %s" % choice_name)

        self.choice_names.append(choice_name)
        self.redis.hset(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "choices", escape.json_encode(self.choice_names))
        self.refresh()

    def remove_choice(self, choice_name):
        """Adds a choice for the experiment"""

        self.choice_names.remove(choice_name)
        self.redis.hset(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "choices", escape.json_encode(self.choice_names))
        self.refresh()

    def add_play(self, choice, count=1):
        """Increments the play count for a given experiment choice"""
        self.redis.hincrby(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "%s:plays" % choice, count)
        self._choices = None

    def add_reward(self, choice, count=1):
        """Increments the reward count for a given experiment choice"""
        self.redis.hincrby(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "%s:rewards" % choice, count)
        self._choices = None

    def compute_default_choice(self):
        """Computes and sets the default choice"""

        choices = self.choices

        if len(choices) == 0:
            return None

        high_choice = max(choices, key=lambda choice: choice.performance)
        self.redis.hset(EXPERIMENT_REDIS_KEY_TEMPLATE % self.name, "default-choice", high_choice.name)
        self.refresh()
        return high_choice

class ExperimentChoice(object):
    """Represents an experiment choice"""

    def __init__(self, experiment, name):
        self.experiment = experiment
        self.name = name
        self.refresh()

    def refresh(self):
        """Re-pulls the data from redis"""

        redis_key = EXPERIMENT_REDIS_KEY_TEMPLATE % self.experiment.name
        self.plays = int(self.experiment.redis.hget(redis_key, "%s:plays" % self.name) or 0)
        self.rewards = int(self.experiment.redis.hget(redis_key, "%s:rewards" % self.name) or 0)
        self.performance = float(self.rewards) / max(self.plays, 1)

def add_experiment(redis, name):
    """Adds a new experiment"""

    if not ALLOWED_NAMES.match(name):
        raise ExperimentException(name, "Illegal name")
    if redis.exists(EXPERIMENT_REDIS_KEY_TEMPLATE % name):
        raise ExperimentException(name, "Already exists")

    json = dict(creation_date=util.unicode_type(datetime.datetime.now()))
    pipe = redis.pipeline(transaction=True)
    pipe.sadd(ACTIVE_EXPERIMENTS_REDIS_KEY, name)
    pipe.hset(EXPERIMENT_REDIS_KEY_TEMPLATE % name, "metadata", escape.json_encode(json))
    pipe.execute()
    return Experiment(redis, name)

def get_experiments(redis, active=True):
    """Gets the full list of experiments"""

    key = ACTIVE_EXPERIMENTS_REDIS_KEY if active else ARCHIVED_EXPERIMENTS_REDIS_KEY
    return [Experiment(redis, escape.to_unicode(name)) for name in redis.smembers(key)]



"""Options for the JSON API plugin"""

from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import oz
oz.option("allow_jsonp", type=bool, default=True, help="Whether to allow for JSONP requests")
